# coding: utf-8
from __future__ import unicode_literals

from requests.auth import AuthBase

from .compat import get_bytes

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Text  # noqa
        from requests import Request  # noqa
except ImportError:
    pass


class TokenAuth(AuthBase):
    """ Attaches Nuxeo Token Authentication to the given Request object. """

    def __init__(self, token):
        # type: (Text) -> None
        self.token = token

    def __eq__(self, other):
        # type: (object) -> bool
        return self.token == getattr(other, "token", None)

    def __ne__(self, other):
        # type: (object) -> bool
        return not self == other

    def __call__(self, r):
        # type: (Request) -> Request
        r.headers[get_bytes("X-Authentication-Token")] = self.token
        return r



# coding: utf-8
from __future__ import unicode_literals

from .endpoint import APIEndpoint
from .models import User

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Dict, Optional, Text  # noqa
        from .client import NuxeoClient  # noqa
except ImportError:
    pass


class API(APIEndpoint):
    """ Endpoint for users. """

    def __init__(self, client, endpoint="user", headers=None):
        # type: (NuxeoClient, Text, Optional[Dict[Text, Text]]) -> None
        super(API, self).__init__(client, endpoint=endpoint, cls=User, headers=headers)

    def get(self, user_id=None):
        # type: (Optional[Text]) -> User
        """
        Get the detail of a user.

        :param user_id: the id of the user
        :return: the user
        """
        return super(API, self).get(path=user_id)

    def post(self, user):
        # type: (User) -> User
        """
        Create a user.

        :param user: the user to create
        :return: the created user
        """
        return super(API, self).post(user)

    create = post  # Alias for clarity

    def put(self, user):
        # type: (User) -> User
        """
        Update a user.

        :param user: the user to update
        :return: the updated user
        """
        return super(API, self).put(user)

    def delete(self, user_id):
        # type: (Text) -> None
        """
        Delete a user.

        :param user_id: the id of the user to delete
        """
        super(API, self).delete(user_id)

    def current_user(self):
        # type: () -> User
        """
        Get the current user details and validate the connection to the server at the same time.

        :return User: user's details
        """
        details = self.client.request("POST", "site/automation/login").json()
        return User(
            extendedGroups=details["groups"],
            id=details["username"],
            isAdministrator=details["isAdministrator"],
        )



# coding: utf-8
"""
The default upload handler.
"""
from __future__ import unicode_literals

from ..compat import get_bytes, quote, text
from ..utils import log_chunk_details

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Any, Callable, Generator, Set, Text, Tuple, Union  # noqa
        from .models import Batch, Blob, BufferBlob, FileBlob  # noqa

        ActualBlob = Union[BufferBlob, FileBlob]  # noqa
except ImportError:
    pass


class Uploader(object):
    """ Helper for uploads """

    chunked = False

    def __init__(self, service, batch, blob, chunk_size, callback=None):
        # type: ("API", Batch, ActualBlob, int, Union[Callable, Tuple[Callable]]) -> None
        self.service = service
        self.batch = batch
        self.blob = blob
        self.chunk_size = chunk_size
        self.headers = service.headers.copy()

        # Several callbacks are accepted
        if callback and isinstance(callback, (tuple, list, set)):
            self.callback = tuple(cb for cb in callback if callable(cb))
        else:
            self.callback = tuple([callback] if callable(callback) else [])

        self.blob.uploadType = "chunked" if self.chunked else "normal"
        self.chunk_count = 1
        self.path = "{}/{}".format(self.batch.batchId, self.batch.upload_idx)
        self.headers.update(
            {
                "Cache-Control": "no-cache",
                "X-File-Name": quote(get_bytes(self.blob.name)),
                "X-File-Size": text(self.blob.size),
                "X-File-Type": self.blob.mimetype,
                "Content-Length": text(self.blob.size),
                "Content-Type": self.blob.mimetype,
            }
        )

    def __repr__(self):
        # type: () -> Text
        return "<{} is_complete={!r}, chunked={!r}, chunk_size={!r}, batch={!r}, blob={!r}>".format(
            type(self).__name__,
            self.is_complete(),
            self.chunked,
            self.chunk_size,
            self.batch,
            self.blob,
        )

    def __str__(self):
        # type: () -> Text
        return repr(self)

    def is_complete(self):
        # type: () -> bool
        """Return True when the upload is completely done."""
        return getattr(self, "_completed", False)

    def process(self, response):
        # type: (Blob) -> None
        self.blob.fileIdx = response.fileIdx
        self.blob.uploadedSize = int(response.uploadedSize)

    def upload(self):
        # type: () -> None
        """ Upload the file. """
        with self.blob as src:
            data = src if self.blob.size else None

            self.process(
                self.service.send_data(
                    self.blob.name, data, self.path, self.chunked, 0, self.headers
                )
            )

            setattr(self, "_completed", True)

            for callback in self.callback:
                callback(self)
        self._update_batch()

    def _update_batch(self):
        # type: () -> None
        """ Add the uploaded blob info to the batch. """
        if self.is_complete():
            # All the parts have been uploaded, update the attributes
            self.blob.batch_id = self.batch.uid
            self.batch.blobs[self.batch.upload_idx] = self.blob
            self.batch.upload_idx += 1


class ChunkUploader(Uploader):
    """ Helper for chunked uploads """

    chunked = True

    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None
        super(ChunkUploader, self).__init__(*args, **kwargs)

        self.chunk_count, self.blob.uploadedChunkIds = self.service.state(
            self.path, self.blob, chunk_size=self.chunk_size
        )
        log_chunk_details(
            self.chunk_count,
            self.chunk_size,
            self.blob.uploadedChunkIds,
            self.blob.size,
        )

        self.blob.chunkCount = self.chunk_count
        self.blob.uploadedSize = min(
            self.blob.size, len(self.blob.uploadedChunkIds) * self.chunk_size
        )

        self.headers.update(
            {"X-Upload-Type": "chunked", "X-Upload-Chunk-Count": text(self.chunk_count)}
        )

        self._to_upload = []
        self._compute_chunks_left()

    def _compute_chunks_left(self):
        # type: () -> None
        """ Compare the set of uploaded chunks with the final list. """
        if self.is_complete():
            return

        to_upload = set(range(self.chunk_count)) - set(self.blob.uploadedChunkIds)
        self._to_upload = sorted(list(to_upload))

    def is_complete(self):
        # type: () -> bool
        """Return True when the upload is completely done."""
        return len(self.blob.uploadedChunkIds) == self.chunk_count

    def iter_upload(self):
        # type: () -> Generator
        """
        Get a generator to upload the file.

        If the `Uploader` has callback(s), they are run after each chunk upload.
        The method will yield after the callbacks step. It yields the uploader
        itself since it contains all relevant data.
        """
        with self.blob as src:
            while self._to_upload:
                # Get the index of a chunk to upload
                index = self._to_upload[0]

                # Seek to the right position
                src.seek(index * self.chunk_size)

                # Read a chunk of data
                data = src.read(self.chunk_size)
                data_len = len(data)

                # Upload it
                self.process(
                    self.service.send_data(
                        self.blob.name,
                        data,
                        self.path,
                        self.chunked,
                        index,
                        self.headers,
                        data_len=data_len,
                    )
                )

                # Now that the part is uploaded, remove it from the list
                self._to_upload.pop(0)

                # keep track of the uploaded data size so far
                self.blob.uploadedSize += data_len

                # If the set of chunks to upload is empty, check whether
                # the server has received all of them.
                if not self._to_upload:
                    self._compute_chunks_left()

                # Call the callback(s), if any
                for callback in self.callback:
                    callback(self)

                # Yield to the upper scope
                yield self

        self._update_batch()

    def process(self, response):
        # type: (Blob) -> None
        self.blob.fileIdx = response.fileIdx
        self.blob.uploadedChunkIds = [int(i) for i in response.uploadedChunkIds]

    def upload(self):
        # type: () -> None
        """Helper to upload the file in one-shot."""
        list(self.iter_upload())



# coding: utf-8
from __future__ import unicode_literals

from requests.exceptions import RetryError

from .compat import text

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Any, Dict, List, Optional, Text  # noqa
except ImportError:
    pass


class NuxeoError(Exception):
    """ Mother class for all exceptions. """


class BadQuery(NuxeoError):
    """
    Exception thrown either by an operation failure:
        - when the command is not valid
        - on unexpected parameter
        - on missing required parameter
        - when a parameter has not the required type
    Or by any requests to the server with invalid data.
    """


class CorruptedFile(NuxeoError):
    """ Exception thrown when digests of a downloaded blob are different. """

    def __init__(self, filename, server_digest, local_digest):
        # type: (Text, Text, Text) -> None
        self.filename = filename
        self.server_digest = server_digest
        self.local_digest = local_digest

    def __repr__(self):
        # type: () -> Text
        err = "CorruptedFile {!r}: server digest is {!r}, local digest is {!r}"
        return err.format(self.filename, self.server_digest, self.local_digest)

    def __str__(self):
        # type: () -> Text
        return repr(self)


class HTTPError(RetryError, NuxeoError):
    """ Exception thrown when the server returns an error. """

    _valid_properties = {"status": -1, "message": None, "stacktrace": None}

    def __init__(self, **kwargs):
        # type: (Any) -> None
        for key, default in HTTPError._valid_properties.items():
            if key == "status":
                # Use the subclass value, if defined
                value = getattr(self, key, kwargs.get(key, default))
            else:
                value = kwargs.get(key, default)

            setattr(self, key, value)

    def __repr__(self):
        # type: () -> Text
        return "%s(%d), error: %r, server trace: %r" % (
            type(self).__name__,
            self.status,
            self.message,
            self.stacktrace,
        )

    def __str__(self):
        # type: () -> Text
        return repr(self)

    @classmethod
    def parse(cls, json):
        # type: (Dict[Text, Any]) -> HTTPError
        """ Parse a JSON object into a model instance. """
        model = cls()

        for key, val in json.items():
            if key in cls._valid_properties:
                setattr(model, key, val)
        return model


class Forbidden(HTTPError):
    """ Exception thrown when the HTTPError code is 403. """

    status = 403


class InvalidBatch(NuxeoError):
    """ Exception thrown when accessing inexistant or deleted batches. """


class InvalidUploadHandler(NuxeoError):
    """ Exception thrown when trying to upload a blob using an invalid handler. """

    def __init__(self, handler, handlers):
        # type: (Text, List[Text]) -> None
        self.handler = handler
        self.handlers = tuple(handlers)

    def __repr__(self):
        # type: () -> Text
        msg = "{}: the upload handler {!r} is not one of {}."
        return msg.format(type(self).__name__, self.handler, self.handlers)

    def __str__(self):
        # type: () -> Text
        return repr(self)


class Unauthorized(HTTPError):
    """ Exception thrown when the HTTPError code is 401. """

    status = 401


class UnavailableConvertor(NuxeoError):
    """
    Exception thrown when a converter is registered but not
    available right now (e.g. not installed on the server).

    :param options: options passed to the conversion request
    """

    def __init__(self, options):
        # type: (Dict[Text, Any]) -> None
        self.options = options
        self.message = text(self)

    def __repr__(self):
        # type: () -> Text
        return (
            "UnavailableConvertor: conversion with options {!r} is not available"
        ).format(self.options)

    def __str__(self):
        # type: () -> Text
        return repr(self)


class UploadError(NuxeoError):
    """
    Exception thrown when an upload fails even after retries.
    """

    def __init__(self, name, chunk=None, info=None):
        # type: (Text, Optional[int], Optional[str]) -> None
        self.name = name
        self.chunk = chunk
        self.info = info

    def __repr__(self):
        # type: () -> Text
        err = "UploadError: unable to upload file {!r}".format(self.name)
        if self.chunk:
            err = "{} (failed at chunk {})".format(err, self.chunk)
        if self.info:
            err = "{} (source: {})".format(err, self.info)
        return err

    def __str__(self):
        # type: () -> Text
        return repr(self)



# coding: utf-8
from __future__ import unicode_literals

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Text, Type, Union  # noqa
        from requests import HTTPError  # noqa
except ImportError:
    pass

try:
    from urllib.parse import quote, urlencode
except ImportError:
    from urllib2 import quote  # noqa
    from urllib import urlencode  # noqa

try:
    long = long
except NameError:
    long = int

try:
    text = unicode  # type: Type[Text]
except NameError:
    text = str  # type: Type[Text]


def get_bytes(data):
    # type: (Union[Text, bytes]) -> bytes
    """
    If data is not bytes, encode it.

    :param data: the input data
    :return: the bytes of data
    """
    if not isinstance(data, bytes):
        data = data.encode("utf-8")
    return data


def get_text(data):
    # type: (Union[Text, bytes]) -> Text
    """
    If data is not text, decode it.

    :param data: the input data
    :return: data in unicode
    """
    if not isinstance(data, text):
        data = data.decode("utf-8")
    return data



# coding: utf-8
from __future__ import unicode_literals

# Force parameters verification for all operations
CHECK_PARAMS = False

# Chunk size to download files
CHUNK_SIZE = 8192  # 8 KiB

# API paths
DEFAULT_API_PATH = "api/v1"
DEFAULT_URL = "http://localhost:8080/nuxeo/"

# Default value for the:
#   - 'X-Application-Name' HTTP header
#   - 'applicationName' URL parameter
DEFAULT_APP_NAME = "Python client"

# Retries for each HTTP call on conection error
MAX_RETRY = 5

# Backoff factor between each retry
# Ex: with 0.2 then sleep() will sleep for [0.2s, 0.4s, 0.8s, ...] between retries
# Ex: with 1 then sleep() will sleep for [1s, 2s, 4s, ...] between retries
RETRY_BACKOFF_FACTOR = 1

# HTTP methods we want to handle in retries
RETRY_METHODS = frozenset(["GET", "POST", "PUT", "DELETE"])

# HTTP status code to handle for retries
RETRY_STATUS_CODES = [429, 500, 503, 504]

# Connection and read timeout, in seconds
TIMEOUT_CONNECT = 5
TIMEOUT_READ = 30

# Size of chunks for the upload
UPLOAD_CHUNK_SIZE = 20 * 1024 * 1024  # 20 MiB

# Upload providers
UP_AMAZON_S3 = "s3"



# coding: utf-8
from __future__ import unicode_literals

from .endpoint import APIEndpoint
from .models import Comment
from .utils import SwapAttr

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Any, Dict, List, Optional, Text  # noqa
        from .client import NuxeoClient  # noqa
except ImportError:
    pass


class API(APIEndpoint):
    """ Endpoint for comments. """

    def __init__(self, client, endpoint="@comment", headers=None):
        # type: (NuxeoClient, Text, Optional[Dict[Text, Text]]) -> None
        super(API, self).__init__(
            client, endpoint=endpoint, cls=Comment, headers=headers
        )

    def get(self, uid):
        # type: (Text) -> Comment
        """
        Get the detail of a comment.

        :param uid: the ID of the comment
        :return: the comment
        """
        # Adding "&fetch-comment=repliesSummary" to the URL to retrieve replies number as well
        return super(API, self).get(
            path=uid, params={"fetch-comment": "repliesSummary"}
        )

    def post(self, comment):
        # type: (Comment) -> Comment
        """
        Create a comment.

        :param comment: the comment to create
        :return: the created comment
        """
        return super(API, self).post(comment)

    create = post  # Alias for clarity

    def put(self, resource):
        # type: (Comment) -> None
        """
        Update a comment.

        :param resource: the entry to update
        :return: the entry updated
        """
        return super(API, self).put(resource)

    def delete(self, uid):
        # type: (Text) -> None
        """
        Delete a comment.

        :param uid: the ID of the comment to delete
        """
        super(API, self).delete(uid)

    def replies(self, uid, params=None):
        # type: (Text, Any) -> List[Comment]
        """
        Get the replies of the comment.

        Any additionnal arguments will be passed to the *params* parent's call.

        :param uid: the ID of the comment
        :return: the list of replies
        """
        # Adding "&fetch-comment=repliesSummary" to the URL to retrieve replies number as well
        kwargs = {"fetch-comment": "repliesSummary"}
        if isinstance(params, dict):
            kwargs.update(params)

        endpoint = "{}/id/{}/@comment".format(self.client.api_path, uid)

        with SwapAttr(self, "endpoint", endpoint):
            return super(API, self).get(params=kwargs)



# coding: utf-8
from __future__ import unicode_literals

import atexit
import json
import logging
import sys
from warnings import warn

from urllib3.util.retry import Retry

import requests
from requests.adapters import HTTPAdapter

from . import (
    __version__,
    comments,
    directories,
    documents,
    groups,
    operations,
    tasks,
    uploads,
    users,
    workflows,
)
from .auth import TokenAuth
from .compat import text
from .constants import (
    CHUNK_SIZE,
    DEFAULT_API_PATH,
    DEFAULT_APP_NAME,
    DEFAULT_URL,
    MAX_RETRY,
    RETRY_BACKOFF_FACTOR,
    RETRY_METHODS,
    RETRY_STATUS_CODES,
    TIMEOUT_CONNECT,
    TIMEOUT_READ,
)
from .exceptions import BadQuery, Forbidden, HTTPError, Unauthorized
from .utils import json_helper

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Any, Dict, Optional, Text, Tuple, Type, Union  # noqa
        from requests.auth import AuthBase  # noqa

        AuthType = Optional[Union[Tuple[Text, Text], AuthBase]]
except ImportError:
    pass

logger = logging.getLogger(__name__)


# Custom exception to raise based on the HTTP status code
# (default will be HTTPError)
HTTP_ERROR = {
    requests.codes.forbidden: Forbidden,
    requests.codes.unauthorized: Unauthorized,
}


class NuxeoClient(object):
    """
    The HTTP client used by Nuxeo.

    :param auth: An authentication object passed to Requests
    :param host: The url of the Nuxeo Platform
    :param api_path: The API path appended to the host url
    :param chunk_size: The size of the chunks for blob download
    :param kwargs: kwargs passed to :func:`NuxeoClient.request`
    """

    def __init__(
        self,
        auth=None,  # type: AuthType
        host=DEFAULT_URL,  # type: Text
        api_path=DEFAULT_API_PATH,  # type: Text
        chunk_size=CHUNK_SIZE,  # type: int
        **kwargs  # type: Any
    ):
        # type: (...) -> None
        self.auth = auth
        self.host = host
        self.api_path = api_path
        self.chunk_size = chunk_size

        version = kwargs.pop("version", "")
        app_name = kwargs.pop("app_name", DEFAULT_APP_NAME)
        self.headers = {
            "X-Application-Name": app_name,
            "X-Client-Version": version,
            "User-Agent": app_name + "/" + version,
            "Accept": "application/json, */*",
        }
        self.schemas = kwargs.get("schemas", "*")
        self.repository = kwargs.pop("repository", "default")
        self._session = requests.sessions.Session()
        cookies = kwargs.pop("cookies", None)
        if cookies:
            self._session.cookies = cookies
        self._session.stream = True
        self.client_kwargs = kwargs
        atexit.register(self.on_exit)

        # Cache for the server information
        self._server_info = None

        # Ensure the host is well formatted
        if not self.host.endswith("/"):
            self.host += "/"

        # The retry adapter
        self.retries = Retry(
            total=MAX_RETRY,
            backoff_factor=RETRY_BACKOFF_FACTOR,
            method_whitelist=RETRY_METHODS,
            status_forcelist=RETRY_STATUS_CODES,
        )

        # Install the retries mecanism
        self.enable_retry()

    def __repr__(self):
        # type: () -> Text
        fmt = "{name}<host={cls.host!r}, version={cls.server_version!r}>"
        return fmt.format(name=type(self).__name__, cls=self)

    def __str__(self):
        # type: () -> Text
        return repr(self)

    def on_exit(self):
        # type: () -> None
        self._session.close()

    def enable_retry(self):
        # type: () -> None
        """ Set a max retry for all connection errors with an adaptative backoff. """
        self._session.mount("https://", HTTPAdapter(max_retries=self.retries))
        self._session.mount("http://", HTTPAdapter(max_retries=self.retries))

    def disable_retry(self):
        # type: () -> None
        """
        Restore default mount points to disable the eventual retry
        adapters set with .enable_retry().
        """
        self._session.mount("https://", HTTPAdapter())
        self._session.mount("http://", HTTPAdapter())

    def query(
        self,
        query,  # type: Text
        params=None,  # type: Dict[Text, Any]
    ):
        """
        Query the server with the specified NXQL query.
        Additional qery parameters can be set via the `params` argument:

            >>> nuxeo.client.query('NXSQL query', params={'properties': '*'})

        You can find what parameters to tweak under the `Repository.Query`
        operation details.
        """

        data = {"query": query}
        if params:
            data.update(params)

        url = self.api_path + "/search/lang/NXQL/execute"
        return self.request("GET", url, params=data).json()

    def set(self, repository=None, schemas=None):
        # type: (Optional[Text], Optional[Text]) -> NuxeoClient
        """
        Set the repository and/or the schemas for the requests.

        :return: The client instance after adding the settings
        """
        if repository:
            self.repository = repository

        if schemas:
            if isinstance(schemas, list):
                schemas = ",".join(schemas)
            self.schemas = schemas

        return self

    def request(
        self,
        method,  # type: Text
        path,  # type: Text
        headers=None,  # type: Optional[Dict[Text, Text]]
        data=None,  # type: Optional[Any]
        raw=False,  # type: bool
        **kwargs  # type: Any
    ):
        # type: (...) -> Union[requests.Response, Any]
        """
        Send a request to the Nuxeo server.

        :param method: the HTTP method
        :param path: the path to append to the host
        :param headers: the headers for the HTTP request
        :param data: data to put in the body
        :param raw: if True, don't parse the data to JSON
        :param kwargs: other parameters accepted by
               :func:`requests.request`
        :return: the HTTP response
        """
        if method not in (
            "GET",
            "HEAD",
            "POST",
            "PUT",
            "DELETE",
            "CONNECT",
            "OPTIONS",
            "TRACE",
        ):
            raise BadQuery("method parameter is not a valid HTTP method.")

        # Construct the full URL without double slashes
        url = self.host + path.lstrip("/")
        if "adapter" in kwargs:
            url = "{}/@{}".format(url, kwargs.pop("adapter"))

        kwargs.update(self.client_kwargs)

        # Set the default value to `object` to allow someone
        # to set `timeout` to `None`.
        if kwargs.get("timeout", object) is object:
            kwargs["timeout"] = (TIMEOUT_CONNECT, TIMEOUT_READ)

        headers = headers or {}
        if "Content-Type" not in headers:
            headers["Content-Type"] = kwargs.pop("content_type", "application/json")
        headers.update(
            {"X-NXDocumentProperties": self.schemas, "X-NXRepository": self.repository}
        )
        enrichers = kwargs.pop("enrichers", None)
        if enrichers:
            headers["enrichers-document"] = ", ".join(enrichers)

        headers.update(self.headers)
        self._check_headers_and_params_format(headers, kwargs.get("params") or {})

        if data and not isinstance(data, bytes) and not raw:
            data = json.dumps(data, default=json_helper)

        # Set the default value to `object` to allow someone
        # to set `default` to `None`.
        default = kwargs.pop("default", object)

        logger.debug(
            ("Calling {!r} with headers={!r}, params={!r} and cookies={!r}").format(
                url,
                headers,
                kwargs.get("params", data if not raw else {}),
                self._session.cookies,
            )
        )

        exc = None
        try:
            resp = self._session.request(
                method, url, headers=headers, auth=self.auth, data=data, **kwargs
            )
            resp.raise_for_status()
        except Exception as exc:
            if default is object:
                raise self._handle_error(exc)
            resp = default
        else:
            self._log_response(resp, self.chunk_size)
        finally:
            # Explicitly break a reference cycle
            exc = None
            del exc

        return resp

    def _check_headers_and_params_format(self, headers, params):
        # type: (Dict[text, Any], Dict[text, Any]) -> None
        """Check headers and params keys for dots or underscores and throw a warning if one is found."""

        msg = "{!r} {} should not contain '_' nor '.'. Replace with '-' to get rid of that warning."

        for key in headers.keys():
            if "_" in key or "." in key:
                warn(msg.format(key, "header"), DeprecationWarning, 2)

        if not isinstance(params, dict):
            return
        for key in params.keys():
            if "_" in key or "." in key:
                warn(msg.format(key, "param"), DeprecationWarning, 2)

    def request_auth_token(
        self,
        device_id,  # type: Text
        permission,  # type: Text
        app_name=DEFAULT_APP_NAME,  # type: Text
        device=None,  # type: Optional[Text]
        revoke=False,  # type: bool
    ):
        # type: (...) -> Text
        """
        Request a token for the user.

        :param device_id: device identifier
        :param permission: read/write permissions
        :param app_name: application name
        :param device: optional device description
        :param revoke: revoke the token
        """

        parameters = {
            "deviceId": device_id,
            "applicationName": app_name,
            "permission": permission,
            "revoke": text(revoke).lower(),
        }
        if device:
            parameters["deviceDescription"] = device

        path = "authentication/token"
        token = self.request("GET", path, params=parameters).text

        # Use the (potentially re-newed) token from now on
        if not revoke:
            self.auth = TokenAuth(token)
        return token

    def is_reachable(self):
        # type: () -> bool
        """ Check if the Nuxeo Platform is reachable. """
        response = self.request("GET", "runningstatus", default=False)
        if isinstance(response, requests.Response):
            return response.ok
        else:
            return bool(response)

    def server_info(self, force=False):
        # type: (bool) -> Dict[Text, Text]
        """
        Retreive server information.

        :param bool force: Force information renewal.
        """

        if force or self._server_info is None:
            response = self.request("GET", "json/cmis", default={})
            if isinstance(response, requests.Response):
                info = response.json()["default"]
            else:
                info = response
            self._server_info = info
        return self._server_info

    @property
    def server_version(self):
        # type: () -> Text
        """ Return the server version. """
        return self.server_info().get("productVersion", "")

    @staticmethod
    def _handle_error(error):
        # type: (Exception) -> Exception
        """
        Log error and handle raise.

        :param error: The error to handle
        """
        if isinstance(error, requests.HTTPError):
            try:
                error_data = error.response.json()
            except ValueError:
                error_data = {
                    "status": error.response.status_code,
                    "message": error.response.content,
                }

            error_cls = HTTP_ERROR.get(error_data.get("status"), HTTPError)
            error = error_cls.parse(error_data)
        return error

    @staticmethod
    def _log_response(response, limit_size):
        # type: (requests.Response, int) -> None
        """
        Log the server's response based on its content type.

        :param response: The server's response to handle
        :param limit_size: Maximum size to not overflow when printing raw content
        of the response
        """

        # No need to do more work if nobody will see it
        if logger.getEffectiveLevel() > logging.DEBUG:
            return

        headers = response.headers
        content_type = headers.get("content-type", "application/octet-stream")
        content_size = int(headers.get("content-length", 0))

        if content_size <= 0:
            # response.content is empty when *void_op* is True,
            # meaning we do not want to get back what we sent
            # or the operation does not return anything by default
            content = "<no content>"
        elif not content_type.startswith("text") and "json" not in content_type:
            # The Content-Type is a binary one, but it does not contain JSON data
            # Skipped binary types are everything but "text/xxx":
            #   https://www.iana.org/assignments/media-types/media-types.xhtml
            content = "<binary data ({:,} bytes)>".format(content_size)
        elif response.url.endswith("site/automation"):
            # This endpoint returns too many information and pollute logs.
            # Besides contents of this call are stored into the .operations attr.
            content = "<Automation details saved into the *operations* attr>"
        elif response.url.endswith("json/cmis"):
            # This endpoint returns too many information and pollute logs.
            # Besides contents of this call are stored into the .server_info attr.
            content = "<CMIS details saved into the *server_info* attr>"
        elif content_size <= limit_size:
            # At this point, we should only have text data not bigger than *limit_size*.
            # Do not use response.text as it will load the chardet module and its
            # heavy encoding detection mecanism. The server will only return UTF-8.
            # See https://stackoverflow.com/a/24656254/1117028 and NXPY-100.
            try:
                content = response.content.decode("utf-8", errors="replace")
            except (MemoryError, OverflowError):
                # OverflowError: join() result is too long (in requests.models.content)
                # Still check for memory errors, this should never happen; or else,
                # it should be painless.
                content = "<no enough memory ({:,} bytes)>".format(content_size)
        else:
            # Here too, at this point, we should only have text data, but too much to display
            # (maybe worth to remove that limit?)
            content = "<too much text ({:,} bytes)>".format(content_size)

        logger.debug(
            "Response from {!r}: {!r} with headers {!r} and cookies {!r}".format(
                response.url, content, headers, response.cookies
            )
        )


class Nuxeo(object):
    """
    Instantiate the client and all the API Endpoints.

    :param auth: the authenticator
    :param host: the host URL
    :param app_name: the name of the application using the client
    :param client: the client class
    :param kwargs: any other argument to forward to every requests calls
    """

    def __init__(
        self,
        auth=None,  # type: Optional[Tuple[Text, Text]]
        host=DEFAULT_URL,  # type: Text
        app_name=DEFAULT_APP_NAME,  # type: Text
        version=__version__,  # type: Text
        client=NuxeoClient,  # type: Type[NuxeoClient]
        **kwargs  # type: Any
    ):
        # type: (...) -> None

        if sys.version_info[:2] == (2, 7):  # pragma: no cover
            from warnings import warn

            message = (
                "Python 2.7 will reach the end of its life on January"
                " 1st, 2020. Please upgrade your Python as Python 2.7"
                " won't be maintained after that date."
                " And so the Nuxeo Python client."
            )
            warn(message, DeprecationWarning, 2)

        self.client = client(
            auth, host=host, app_name=app_name, version=version, **kwargs
        )
        self.comments = comments.API(self.client)
        self.operations = operations.API(self.client)
        self.directories = directories.API(self.client)
        self.groups = groups.API(self.client)
        self.tasks = tasks.API(self.client)
        self.uploads = uploads.API(self.client)
        self.users = users.API(self.client)
        self.workflows = workflows.API(self.client, self.tasks)
        self.documents = documents.API(
            self.client, self.operations, self.workflows, self.comments
        )



# coding: utf-8
from __future__ import unicode_literals
from os import fsync

try:
    from collections.abc import Sequence
except ImportError:
    from collections import Sequence

from . import constants
from .compat import get_text, long, text
from .endpoint import APIEndpoint
from .exceptions import BadQuery, CorruptedFile
from .models import Blob, Operation
from .utils import get_digester

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from requests import Response  # noqa
        from typing import Any, Dict, Optional, Text, Tuple, Type  # noqa
        from .client import NuxeoClient  # noqa
except ImportError:
    pass

# Types allowed for operations parameters
# See https://docs.oracle.com/javase/tutorial/java/nutsandbolts/datatypes.html
# for default values
PARAM_TYPES = {
    # Operation: ((accepted types), default value if optional))
    "blob": ((text, bytes, Blob), None),
    "boolean": ((bool,), False),
    "date": ((text, bytes), None),
    "document": ((text, bytes), None),
    "documents": ((list,), None),
    "int": ((int,), 0),
    "integer": ((int,), 0),
    "long": ((int, long), 0),
    "list": ((list,), None),
    "map": ((dict,), None),
    "object": ((object,), None),
    "properties": ((dict,), None),
    "resource": ((text, bytes), None),
    "serializable": ((Sequence,), None),
    "string": ((text, bytes), None),
    "stringlist": ((Sequence,), None),
    "validationmethod": ((text, bytes), None),
}  # type: Dict[Text, Tuple[Type, ...]]


class API(APIEndpoint):
    """ Endpoint for operations. """

    # Operations cache
    ops = {}  # type: Dict[Text, Any]

    def __init__(self, client, endpoint="site/automation", headers=None):
        # type: (NuxeoClient, Text, Optional[Dict[Text, Text]]) -> None
        headers = headers or {}
        headers.update({"Content-Type": "application/json", "X-NXproperties": "*"})
        super(API, self).__init__(client, endpoint=endpoint, cls=dict, headers=headers)
        self.endpoint = endpoint

    def get(self, **kwargs):
        # type: (Any) -> Dict[Text, Any]
        """ Get the list of available operations from the server. """
        return super(API, self).get()

    def put(self, **kwargs):
        # type: (Any) -> None
        raise NotImplementedError()

    def delete(self, resource_id):
        # type: (Text) -> None
        raise NotImplementedError()

    @property
    def operations(self):
        # type: () -> Dict[Text, Any]
        """
        Get a dict of available operations.

        :return: the available operations
        """
        if not self.ops:
            response = self.get()
            for operation in response["operations"]:
                self.ops[operation["id"]] = operation
                for alias in operation.get("aliases", []):
                    self.ops[alias] = operation

        return self.ops

    def check_params(self, command, params):
        # type: (Text, Dict[Text, Any]) -> None
        """
        Check given parameters of the `command` operation.  It will also
        check for types whenever possible.
        """

        operation = self.operations.get(command)
        if not operation:
            raise BadQuery("{!r} is not a registered operation".format(command))

        parameters = {param["name"]: param for param in operation["params"]}

        for name, value in params.items():
            # Check for unexpected parameters.  We use `dict.pop()` to
            # get and delete the parameter from the dict.
            try:
                param = parameters.pop(name)
            except KeyError:
                err = "unexpected parameter {!r} for operation {}"
                raise BadQuery(err.format(name, command))

            # Check types
            types_accepted, default = PARAM_TYPES[param["type"]]
            if not param["required"]:
                # Allow the default value when the parameter is not required
                types_accepted += (type(default),)
            types_accepted = tuple(set(types_accepted))  # Uniquify

            if not isinstance(value, types_accepted):
                types = [type_.__name__ for type_ in types_accepted]
                if len(types) > 1:
                    types = ", ".join(types[:-1]) + " or " + types[-1]
                else:
                    types = types[0]
                err = "parameter {}={!r} should be of type {} (current is {})"
                raise BadQuery(err.format(name, value, types, type(value).__name__))

        # Check for required parameters.  As of now, `parameters` may contain
        # unclaimed parameters and we just need to check for required ones.
        for (name, parameter) in parameters.items():
            if parameter["required"]:
                err = "missing required parameter {!r} for operation {!r}"
                raise BadQuery(err.format(name, command))

    def execute(
        self,
        operation=None,  # type: Optional[Operation]
        void_op=False,  # type: bool
        headers=None,  # type: Optional[Dict[Text, Text]]
        file_out=None,  # type: Optional[Text]
        **kwargs  # type: Any
    ):
        # type: (...) -> Any
        """
        Execute an operation.

        If there is no operation parameter, the command,
        the input object, and the parameters of the operation
        will be taken from the kwargs.

        :param operation: the operation
        :param void_op: if True, the body of the response
        from the server will be empty
        :param headers: extra HTTP headers
        :param file_out: if not None, path of the file
        where the response will be saved
        :param kwargs: any other parameter
          *callback* is either a single callable or a tuple of callables.
        :return: the result of the execution
        """
        json = kwargs.pop("json", True)
        callback = kwargs.pop("callback", None)
        enrichers = kwargs.pop("enrichers", None)
        check_params = kwargs.pop("check_params", constants.CHECK_PARAMS)
        default = kwargs.pop("default", object)
        timeout = kwargs.pop("timeout", object)

        command, input_obj, params, context = self.get_attributes(operation, **kwargs)

        if check_params:
            self.check_params(command, params)

        url = "site/automation/{}".format(command)
        if isinstance(input_obj, Blob):
            url = "{}/upload/{}/{}/execute/{}".format(
                self.client.api_path, input_obj.batch_id, input_obj.fileIdx, command
            )
            input_obj = None

        headers = headers or {}
        headers.update(self.headers)
        if void_op:
            headers["X-NXVoidOperation"] = "true"

        data = self.build_payload(params, context)

        if input_obj:
            if isinstance(input_obj, list):
                input_obj = "docs:" + ",".join(input_obj)
            data["input"] = input_obj

        resp = self.client.request(
            "POST",
            url,
            data=data,
            headers=headers,
            enrichers=enrichers,
            default=default,
            timeout=timeout,
        )

        # Save to a file, part by part of chunk_size
        if file_out:
            return self.save_to_file(
                operation, resp, file_out, callback=callback, **kwargs
            )

        # It is likely a JSON response we do not want to save to a file
        if operation:
            operation.progress = int(resp.headers.get("content-length", 0))

        if json:
            try:
                return resp.json()
            except ValueError:
                pass
        return resp.content

    @staticmethod
    def get_attributes(operation, **kwargs):
        # type: (Operation, Any) -> (Text, Any, Dict[Text, Any])
        """ Get the operation attributes. """
        if operation:
            command = operation.command
            input_obj = operation.input_obj
            context = operation.context
            params = operation.params
        else:
            command = kwargs.pop("command", None)
            input_obj = kwargs.pop("input_obj", None)
            context = kwargs.pop("context", None)
            params = kwargs.pop("params", kwargs)
        return command, input_obj, params, context

    def build_payload(self, params, context):
        # type: (Dict[Text, Any], Dict[Text, Any]) -> Dict[Text, Any]
        """ Create sanitized operation payload. """
        data = {"params": self.sanitize(params)}  # type: Dict[Text, Any]
        clean_context = self.sanitize(context)  # type: Dict[Text, Any]
        if clean_context:
            data["context"] = clean_context

        return data

    @staticmethod
    def sanitize(obj):
        # type: (Dict[Text, Any]) -> Dict[Text, Any]
        """ Sanitize the operation parameters. """
        if not obj:
            return {}

        clean_obj = {}  # type: Dict[Text, Any]

        for k, v in obj.items():
            if v is None:
                continue

            if k != "properties":
                clean_obj[k] = v
                continue

            # v can only be a dict
            contents = [
                "{}={}".format(name, get_text(value)) for name, value in v.items()
            ]
            clean_obj[k] = "\n".join(contents)

        return clean_obj

    def new(self, command, **kwargs):
        # type: (Text, Any) -> Operation
        """ Make a new Operation object. """
        return Operation(command=command, service=self, **kwargs)

    def save_to_file(self, operation, resp, path, **kwargs):
        # type: (Operation, Response, Text, Any) -> Text
        """
        Save the result of an operation to a file.

        If there is a digest of the file to check
        against the server, it can be passed in
        the kwargs.
        :param operation: the operation
        :param resp: the response from the Platform
        :param path: the path to save the file to
        :param kwargs: additional parameters
        :return:
        """
        digest = kwargs.pop("digest", None)
        digester = get_digester(digest) if digest else None

        unlock_path = kwargs.pop("unlock_path", None)
        lock_path = kwargs.pop("lock_path", None)
        use_lock = callable(unlock_path) and callable(lock_path)
        locker = None

        # Several callbacks are accepted, tuple is used to keep order
        callback = kwargs.pop("callback", None)
        if callback and isinstance(callback, (tuple, list, set)):
            callbacks = tuple(cb for cb in callback if callable(cb))
        else:
            callbacks = tuple([callback] if callable(callback) else [])

        if use_lock:
            locker = unlock_path(path)
        try:
            with open(path, "ab") as f:
                chunk_size = kwargs.get("chunk_size", self.client.chunk_size)
                for chunk in resp.iter_content(chunk_size=chunk_size):
                    # Check if synchronization thread was suspended
                    for callback in callbacks:
                        callback(path)
                    if operation:
                        operation.progress += chunk_size
                    f.write(chunk)
                    if digester:
                        digester.update(chunk)

                # Force write of file to disk
                f.flush()
                fsync(f.fileno())
        finally:
            if use_lock:
                lock_path(path, locker)

        if digester:
            computed_digest = digester.hexdigest()
            if digest != computed_digest:
                raise CorruptedFile(path, digest, computed_digest)

        return path



# coding: utf-8
from __future__ import unicode_literals

from .endpoint import APIEndpoint
from .models import Group

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Dict, Optional, Text  # noqa
        from .client import NuxeoClient  # noqa
except ImportError:
    pass


class API(APIEndpoint):
    """ Endpoint for groups. """

    def __init__(self, client, endpoint="group", headers=None):
        # type: (NuxeoClient, Text, Optional[Dict[Text, Text]]) -> None
        self.params = {"fetch-group": ["memberUsers", "memberGroups"]}
        super(API, self).__init__(client, endpoint=endpoint, cls=Group, headers=headers)

    def get(self, group_id=None):
        # type: (Optional[Text]) -> Group
        """
        Get the detail of a group.

        :param group_id: the id of the group
        :return: the group
        """
        return super(API, self).get(path=group_id, params=self.params)

    def post(self, group):
        # type: (Group) -> Group
        """
        Create a group.

        :param group: the group to create
        :return: the created group
        """
        return super(API, self).post(resource=group, params=self.params)

    create = post  # Alias for clarity

    def put(self, group):
        # type: (Group) -> Group
        """
        Update a group.

        :param group: the group to update
        :return: the updated group
        """
        return super(API, self).put(group)

    def delete(self, group_id):
        # type: (Text) -> None
        """
        Delete a group.

        :param group_id: the id of the group to delete
        """
        super(API, self).delete(group_id)



# coding: utf-8
from __future__ import unicode_literals

from .endpoint import APIEndpoint
from .models import Workflow
from .utils import SwapAttr

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Any, Dict, List, Optional, Text, Union  # noqa
        from .client import NuxeoClient  # noqa
        from .models import Document, Task  # noqa
        from .tasks import API as TasksAPI  # noqa
except ImportError:
    pass


class API(APIEndpoint):
    """ Endpoint for workflows. """

    def __init__(
        self,
        client,  # type: NuxeoClient
        tasks,  # type: TasksAPI
        endpoint="workflow",  # type: Text
        headers=None,  # type: Optional[Dict[Text, Text]]
    ):
        # type: (...) -> None
        self.tasks_api = tasks
        super(API, self).__init__(
            client, endpoint=endpoint, cls=Workflow, headers=headers
        )

    def get(self, workflow_id=None):
        # type: (Optional[Text]) -> Workflow
        """
        Get the detail of a workflow.

        :param workflow_id: the id of the workflow
        :return: the workflow
        """
        return super(API, self).get(path=workflow_id)

    def post(
        self,
        model,  # type: Text
        document=None,  # type: Optional[Document]
        options=None,  # type: Optional[Dict[Text, Any]]
    ):
        # type: (...) -> Workflow
        """
        Start a workflow.

        :param model: the workflow to start
        :param document: the document to start the workflow on
        :param options: options for the workflow
        :return: the created workflow
        """
        data = {"workflowModelName": model, "entity-type": "workflow"}
        options = options or {}
        if "attachedDocumentIds" in options:
            data["attachedDocumentIds"] = options["attachedDocumentIds"]
        if "variables" in options:
            data["variables"] = options["variables"]

        if document:
            path = "id/{}/@workflow".format(document.uid)
            with SwapAttr(self, "endpoint", self.client.api_path):
                workflow = super(API, self).post(data, path=path)
        else:
            workflow = super(API, self).post(data)
        return workflow

    start = post  # Alias for clarity

    def put(self, **kwargs):
        # type: (Any) -> None
        raise NotImplementedError()

    def delete(self, workflow_id):
        # type: (Text) -> None
        """
        Delete a workflow.

        :param workflow_id: the id of the workflow to delete
        """
        super(API, self).delete(workflow_id)

    def graph(self, workflow):
        # type: (Workflow) -> Dict[Text, Any]
        """
        Get the graph of the workflow in JSON format.

        :param workflow: the worklow to get the graph from
        :return: the graph
        """
        request_path = "{}/graph".format(workflow.uid)
        return super(API, self).get(path=request_path)

    def started(self, model):
        # type: (Text) -> List[Workflow]
        """
        Get started workflows having the specified model.

        :param model: the workflow model
        :return: the started workflows
        """
        return super(API, self).get(params={"workflowModelName": model})

    def tasks(self, workflow):
        # type: (Workflow) -> List[Task]
        """ Get the tasks of a workflow. """
        return self.tasks_api.get(workflow.as_dict())



# coding: utf-8
"""
The Amazon S3 upload handler.
"""
from __future__ import unicode_literals

import logging

import boto3.session
from botocore.client import Config

from .default import Uploader
from ..constants import UP_AMAZON_S3
from ..exceptions import UploadError
from ..utils import chunk_partition, log_chunk_details

try:
    from typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Any, Generator, List, Set, Tuple  # noqa
except ImportError:
    pass


logger = logging.getLogger(__name__)


class UploaderS3(Uploader):
    """ Helper for uploads using Amazon S3 Direct Upload (single). """

    def __init__(self, *args, **kwargs):
        # Allow to pass a custom S3 client (for tests)
        s3_client = kwargs.pop("s3_client", None)

        super(UploaderS3, self).__init__(*args, **kwargs)

        # Instantiate the S3 client
        s3_info = self.batch.extraInfo
        self.bucket = s3_info["bucket"]
        self.key = "{}/{}".format(s3_info["baseKey"].rstrip("/"), self.blob.name)

        if not s3_client:
            # https://boto3.amazonaws.com/v1/documentation/api/latest/guide/resources.html#multithreading-multiprocessing
            session = boto3.session.Session()
            path_style = "path" if s3_info.get("usePathStyleAccess", False) else "auto"
            s3_client = session.client(
                "s3",
                aws_access_key_id=s3_info["awsSecretKeyId"],
                aws_secret_access_key=s3_info["awsSecretAccessKey"],
                aws_session_token=s3_info["awsSessionToken"],
                endpoint_url=s3_info.get("endpoint") or None,
                region_name=s3_info["region"],
                config=Config(s3={"addressing_style": path_style}),
            )
        self.s3_client = s3_client

    def upload(self):
        # type: () -> None
        """ Upload the file. """
        with self.blob as fd:
            try:
                # Note: we are using put_object() rather than upload_fileobj()
                # to be able to retrieve the ETag from the response. The latter
                # returns nothing and it would involve doing another HTTP call
                # just to get that information.
                response = self.s3_client.put_object(
                    Bucket=self.bucket, Key=self.key, Body=fd
                )
            except Exception as e:
                raise UploadError(self.blob.path, info=str(e))

            # Save the ETag for the batch.complete() call
            self.batch.etag = response["ETag"]

        self.blob.uploadedSize = self.blob.size
        setattr(self, "_completed", True)

        for callback in self.callback:
            callback(self)

        self._update_batch()


class ChunkUploaderS3(UploaderS3):
    """ Helper for chunked uploads using Amazon S3 Direct Upload (multipart). """

    chunked = True

    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None

        # Allow to use a custom *MaxParts* value, used in .state() (for tests)
        # 0 <= MaxParts <= 2,147,483,647 (default is 1,000)
        self._max_parts = kwargs.pop("max_parts", 1000)

        super(ChunkUploaderS3, self).__init__(*args, **kwargs)

        # Parts already sent
        self._data_packs = []

        self.chunk_count, self.blob.uploadedChunkIds = self.state()
        log_chunk_details(
            self.chunk_count,
            self.chunk_size,
            self.blob.uploadedChunkIds,
            self.blob.size,
        )

        self.blob.chunkCount = self.chunk_count
        self.blob.uploadedSize = min(
            self.blob.size, len(self.blob.uploadedChunkIds) * self.chunk_size
        )

        self._to_upload = []
        self._compute_chunks_left()

    def new(self):
        """
        Instantiate a new multipart upload.

        :return: the multipart upload ID
        """
        mpu = self.s3_client.create_multipart_upload(Bucket=self.bucket, Key=self.key)
        self.batch.multiPartUploadId = mpu["UploadId"]
        return self.batch.multiPartUploadId

    def _state(self):
        # type: () -> List[int]
        """See .state()."""

        uploaded_chunks = []
        data_packs = []
        chunk_size = 0
        first = True

        # 0 <= PartNumberMarker <= 2,147,483,647
        part_number_marker = 0

        while "there are parts":
            info = self.s3_client.list_parts(
                Bucket=self.bucket,
                Key=self.key,
                UploadId=self.batch.multiPartUploadId,
                PartNumberMarker=part_number_marker,
                MaxParts=self._max_parts,
            )

            # Nothing was uploaded yet
            if "Parts" not in info:
                break

            for part in info["Parts"]:
                # Save the part size based on the first recieved part data
                if first:
                    chunk_size = part["Size"]
                    first = False

                index = part["PartNumber"]
                data_packs.append({"ETag": part["ETag"], "PartNumber": index})
                uploaded_chunks.append(index)

            # No more parts
            if not info["IsTruncated"]:
                break

            # TODO: part not yet tested/covered. Will be done with NXPY-147.
            # Next parts batch will start with that number
            part_number_marker = info["NextPartNumberMarker"]

        return chunk_size, uploaded_chunks, data_packs

    def state(self):
        # type: () -> Tuple[int, List]
        """
        Get the state of a multipart upload.

        If the blob upload has not begun yet, the server
        will return a 404 error, so we initialize the
        different values.
        If the blob upload is incomplete, we return the
        values the server sent us.

        :return: the chunk count and uploaded chunks
        """
        if self.batch.multiPartUploadId:
            self.chunk_size, uploaded_chunks, self._data_packs = self._state()
        else:
            # It's a new upload
            self.new()
            uploaded_chunks = []

        # *chunk_size* is overidden on purpose:
        # S3 has limitations and chunk size & count may be different from initial values
        chunk_count, self.chunk_size = chunk_partition(
            self.blob.size, self.chunk_size, handler=UP_AMAZON_S3
        )

        return chunk_count, uploaded_chunks

    def _compute_chunks_left(self):
        # type: () -> None
        """ Compare the set of uploaded chunks with the final list. """
        if self.is_complete():
            return

        # S3 starts counting at 1
        to_upload = set(range(1, self.chunk_count + 1, 1)) - set(
            self.blob.uploadedChunkIds
        )
        self._to_upload = sorted(list(to_upload))

    def is_complete(self):
        # type: () -> bool
        """Return True when the upload is completely done."""
        return len(self.blob.uploadedChunkIds) == self.chunk_count

    def iter_upload(self):
        # type: () -> Generator
        """Upload a file in chunks.

        If the `Uploader` has callback(s), they are run after each chunk upload.
        The method will yield after the callbacks step. It yields the uploader
        itself since it contains all relevant data.
        """
        with self.blob as fd:
            while self._to_upload:
                # Get the index of a chunk to upload
                part_number = self._to_upload[0]

                # Seek to the right position (S3 starts counting at 1)
                position = (part_number - 1) * self.chunk_size
                fd.seek(position)

                # Read a chunk of data
                data = fd.read(self.chunk_size)
                data_len = len(data)

                try:
                    # Upload it
                    part = self.s3_client.upload_part(
                        UploadId=self.batch.multiPartUploadId,
                        Bucket=self.bucket,
                        Key=self.key,
                        PartNumber=part_number,
                        Body=data,
                        ContentLength=data_len,
                    )
                except Exception as e:
                    raise UploadError(self.blob.path, chunk=part_number, info=str(e))

                # Now that the part is uploaded, remove it from the list
                self._to_upload.pop(0)

                self._data_packs.append(
                    {"ETag": part["ETag"], "PartNumber": part_number}
                )
                self.blob.uploadedChunkIds.append(part_number)
                self.blob.uploadedSize += data_len

                # If the set of chunks to upload is empty, check whether
                # the server has received all of them.
                if not self._to_upload:
                    self._compute_chunks_left()

                # Call the callback(s), if any
                for callback in self.callback:
                    callback(self)

                # Yield to the upper scope
                yield self

        # Complete the upload on the S3 side
        response = self.s3_client.complete_multipart_upload(
            Bucket=self.bucket,
            Key=self.key,
            UploadId=self.batch.multiPartUploadId,
            MultipartUpload={"Parts": self._data_packs},
        )

        # Save the ETag for the batch.complete() call
        self.batch.etag = response["ETag"]

        self._update_batch()

    def upload(self):
        # type: () -> None
        """Helper to upload the file in one-shot."""
        list(self.iter_upload())
